/////////////////////////////////////////////////////////////////////////////

#include <eetypes.h>
#include <eekernel.h>
#include <libpad.h>
#include <sifdev.h>
#include <sifrpc.h>
#include <libcdvd.h>

#include "ps2port.h"

#include "gsvga.h"
#include "ps2pak.h"
#include "timer.h"
#include "control.h"
#include "ps2sdr.h"

#include "borstartup.h"

/////////////////////////////////////////////////////////////////////////////

static char ps2hostdev[256];

int printf_ok = 0;

//
// NOT REENTRANT
//
const char *ps2gethostfilename(const char *filename) {
  static char t[256];
  char *dest;
  const char *src;
  int is_cdrom = (ps2hostdev[0] == 'c' || ps2hostdev[0] == 'C');
  char slashchar = '/';

  dest = t;
  src = ps2hostdev;
  while(*src) {
    char c = *src++;
    if(c == '/' || c == '\\') slashchar = c;
    *dest++ = c;
  }
  src = filename;
  while(*src) {
    char c = *src++;
    if(c == '/' || c == '\\') c = slashchar;
    if(is_cdrom && c >= 'a' && c <= 'z') c += ('A' - 'a');
    *dest++ = c;
  }
  if(is_cdrom) {
    *dest++ = ';';
    *dest++ = '1';
  }
  *dest = 0;

  return t;
}

/////////////////////////////////////////////////////////////////////////////

void ps2loadirx(const char *filename) {
  if(sceSifLoadModule(ps2gethostfilename(filename), 0, NULL) < 0) {
    if(printf_ok) printf("unable to load %s; halting\n", ps2gethostfilename(filename));
    for(;;);
  }
}

/////////////////////////////////////////////////////////////////////////////

#ifndef PS2RELEASE
static int log_started = 0;
#endif

void ps2logprintf(const char *fmt, ...) {
#ifndef PS2RELEASE
  static char buf[2048];
  int fd;
  int mode;
  va_list arglist;

  va_start(arglist, fmt);
  vsprintf(buf, fmt, arglist);
  va_end(arglist);

  if(log_started) {
    mode = SCE_WRONLY;
  } else {
    mode = SCE_WRONLY | SCE_CREAT | SCE_TRUNC;
  }

  fd = sceOpen(ps2gethostfilename("ps2log"), mode, 0);
  if(fd >= 0) {
    sceLseek(fd, 0, SCE_SEEK_END);
    sceWrite(fd, buf, strlen(buf));
    sceClose(fd);
    log_started = 1;
  }
#endif
}

/////////////////////////////////////////////////////////////////////////////

unsigned readlsb32(const unsigned char *src) {
  return
    ((((unsigned)(src[0])) & 0xFF) <<  0) |
    ((((unsigned)(src[1])) & 0xFF) <<  8) |
    ((((unsigned)(src[2])) & 0xFF) << 16) |
    ((((unsigned)(src[3])) & 0xFF) << 24);
}

/////////////////////////////////////////////////////////////////////////////
//
// do the necessary stuff when booting from cdrom
//
static void boot_from_cdrom(int mediamode) {

  sceSifInitRpc(0);
  sceCdInit(SCECdINIT);
  sceCdMmode(mediamode);

  while(sceSifRebootIop(ps2gethostfilename("ioprp260.img")) == 0);
  while(sceSifSyncIop() == 0);

  sceSifInitRpc(0);
  sceFsReset();
  sceSifLoadFileReset();
  sceCdInit(SCECdINIT);
  sceCdMmode(mediamode);

}

/////////////////////////////////////////////////////////////////////////////

void bor_main(int argc, char **argv);

/////////////////////////////////////////////////////////////////////////////

int main(int argc, char **argv) {

#ifndef PS2FORCECD
  int i;
  // get the host device name
  strcpy(ps2hostdev, argv[0]);
  i = strlen(ps2hostdev);
  for(; i > 0; i--) {
    char c = ps2hostdev[i-1];
    if(c == '/' || c == '\\' || c == ':') break;
  }
  ps2hostdev[i] = 0;

  // used for cdrom testing
  // if the exe name starts with 'cd', force cdrom0 as the host device
  if(
    (argv[0][i+0] == 'c' || argv[0][i+0] == 'C') &&
    (argv[0][i+1] == 'd' || argv[0][i+1] == 'D')
  ) {
    strcpy(ps2hostdev, "cdrom0:\\");
  }
#else
  strcpy(ps2hostdev, "cdrom0:\\");
#endif

  // perform necessary stuff for booting off a CD
  if(ps2hostdev[0] == 'c' || ps2hostdev[0] == 'C') {
    boot_from_cdrom(SCECdCD);
  }

  // necessary for a few things
  sceSifInitRpc(0);
  sceSifInitIopHeap();

  ChangeThreadPriority(GetThreadId(), 10);

  gsvga_init();

  borstartup_draw();

  timer_init();

  printf_ok = 1;

  control_ps2init();

  ps2sdr_init();

// determine timer freq
/*
  { unsigned a = timer_gettime();
    for(i = 0; i < 60; i++) gsvga_draw((unsigned char*)main);
    a = timer_gettime() - a;
    printf("elapsed time = %u\n", a);
    for(;;);
  }
*/

//for(i = 0;; i++) { printf("%d\n", i); } // do shit

//  printf("%s\n", argv[0]);

/*
{  unsigned total = 0;
  for(i = 0;; i++) {
    unsigned n = 100000;
    unsigned q = (unsigned)tracemalloc("main", n);
    if(q) {
      total += n;
      printf("allocated %d bytes (%X)\n", total, q);
    } else {
      printf("alloc failed! total bytes %d\n", total);
      for(;;);
    }
  }
}
*/

  // open the pak
  ps2pak_init();

  // run the game
  bor_main(0, 0);

  for(;;) { }
}

/////////////////////////////////////////////////////////////////////////////

