/////////////////////////////////////////////////////////////////////////////
//
// ps2sdr - raw 48khz output using sdr
//
/////////////////////////////////////////////////////////////////////////////

#include <eetypes.h>
#include <eekernel.h>
#include <libsdr.h>
#include <sdrcmd.h>
#include <sifdev.h>

#include "ps2port.h"

#include "ps2sdr.h"

/////////////////////////////////////////////////////////////////////////////

/*
int sfsize = 3687204;

unsigned char *filebuf;

int sfofs;

void my_ps2sdr_fillstereosamples(signed short *dest_l, signed short *dest_r, int n) {

  int i;
  for(i = 0; i < n; i++) {
    signed short l, r;
    l = *((signed short*)(filebuf+sfofs)); sfofs += 2;
    r = *((signed short*)(filebuf+sfofs)); sfofs += 2;
    if(sfofs >= sfsize) sfofs = 0;
    dest_l[i] = l;
    dest_r[i] = r;
  }

}
*/

/////////////////////////////////////////////////////////////////////////////
//
// this is the IOP buffer size (two blocks)
// the EE buffer is only half as big (one block)
// buffersize must be a multiple of 2048
//
#define BUFFERSIZE (2048)

#define BLOCKSIZE ((BUFFERSIZE)/2)
#define BLOCKSAMPLES ((BLOCKSIZE)/4)

static u_int iop_buf_address;
static unsigned char *ee_buf_address;

static ps2sdr_fillstereosamples_t thefillfunction = NULL;

/////////////////////////////////////////////////////////////////////////////

void ps2sdr_setfillfunction(ps2sdr_fillstereosamples_t fillfunction) {
  thefillfunction = fillfunction;
}

/////////////////////////////////////////////////////////////////////////////

unsigned ps2sdr_blocksdone = 0;

static void ps2sdr_transintr(int core, void *common) {
  int i;
  // fill the ee buffer
  if(thefillfunction) {
    for(i = 0; i < BLOCKSIZE; i += 1024) {
      thefillfunction(
        (signed short*)(ee_buf_address+  0),
        (signed short*)(ee_buf_address+512),
        256
      );
    }
  } else {
    memset(ee_buf_address, 0, BLOCKSIZE);
  }
  FlushCache(WRITEBACK_DCACHE);
  sceSdTransToIOP(
    ee_buf_address,
    iop_buf_address + (BLOCKSIZE*(ps2sdr_blocksdone&1)),
    BLOCKSIZE,
    1
  );
  ps2sdr_blocksdone++;
}

/////////////////////////////////////////////////////////////////////////////

unsigned ps2sdr_sample_odometer(void) {
  return ps2sdr_blocksdone * ((BUFFERSIZE/2)/4);
}

/////////////////////////////////////////////////////////////////////////////

// 0 to 32767
void ps2sdr_setvolume(unsigned v) {
  if(v > 0x7FFF) { v = 0x7FFF; }
  sceSdRemote (1, rSdSetParam, SD_P_BVOLL, v);
  sceSdRemote (1, rSdSetParam, SD_P_BVOLR, v);
}

/////////////////////////////////////////////////////////////////////////////
//
// Initialization
//
void ps2sdr_init(void) {

//  printf("initializing ps2sdr\n");

  ps2loadirx("libsd.irx");
  ps2loadirx("sdrdrv.irx");

  sceSdRemoteInit();
  sceSdRemoteCallbackInit(5);

  // allocate iop buffer
  iop_buf_address = (u_int)sceSifAllocSysMemory(0, BUFFERSIZE+64, NULL);
  if(!iop_buf_address) {
    printf("ps2sdr_init: couldn't allocate iop memory; halting\n");
    for(;;);
  }
  // make sure iop buffer is aligned
  while(((unsigned)iop_buf_address) & 63) iop_buf_address++;
  // allocate ee buffer
  ee_buf_address = tracemalloc("ps2sdr_init", BLOCKSIZE+64);
  // make sure ee buffer is aligned
  while(((unsigned)ee_buf_address) & 63) ee_buf_address++;

  // clear ee buffer and then copy it to the iop buffer
  memset(ee_buf_address, 0, BLOCKSIZE);
  FlushCache(WRITEBACK_DCACHE);
  sceSdTransToIOP(ee_buf_address, iop_buf_address          , BLOCKSIZE, 1);
  sceSdTransToIOP(ee_buf_address, iop_buf_address+BLOCKSIZE, BLOCKSIZE, 1);

  // initialize sdr stuff
  sceSdRemote( 1, rSdInit , SD_INIT_COLD );

  sceSdRemote( 1, rSdSetParam, SD_CORE_0 | SD_P_MVOLL , 0x3fff );
  sceSdRemote( 1, rSdSetParam, SD_CORE_0 | SD_P_MVOLR , 0x3fff );
  sceSdRemote( 1, rSdSetParam, SD_CORE_1 | SD_P_MVOLL , 0x3fff );
  sceSdRemote( 1, rSdSetParam, SD_CORE_1 | SD_P_MVOLR , 0x3fff );
  sceSdRemote( 1, rSdSetParam, SD_CORE_0 | SD_P_MMIX , ((~(SD_MMIX_MINEL | SD_MMIX_MINER)) & 0x0ff0));
  sceSdRemote( 1, rSdSetParam, SD_CORE_1 | SD_P_MMIX , ((~(SD_MMIX_MINEL | SD_MMIX_MINER)) & 0x0fff));

  // set interrupt handler
  sceSdRemote (1, rSdSetTransIntrHandler, 0, ps2sdr_transintr, 0);

  // begin transfer
  sceSdRemote (1, rSdBlockTrans, 0, SD_TRANS_MODE_WRITE | SD_BLOCK_LOOP, iop_buf_address, BUFFERSIZE);

  sceSdRemote (1, rSdSetParam, SD_P_BVOLL , 0);
  sceSdRemote (1, rSdSetParam, SD_P_BVOLR , 0);

//  printf("ok\n");

//  for(;;);
}

/////////////////////////////////////////////////////////////////////////////

